import pytest
import cmselemental
import numpy

using_msgpack = pytest.mark.skipif(
    cmselemental.util.which_import("msgpack", return_bool=True) is False,
    reason="Not detecting module msgpack. Install package if necessary and add to envvar PYTHONPATH",
)

using_pyyaml = pytest.mark.skipif(
    cmselemental.util.which_import("yaml", return_bool=True) is False,
    reason="Not detecting module pyyaml. Install package if necessary and add to envvar PYTHONPATH",
)


serialize_extensions = [
    "json",
    "json-ext",
    pytest.param("msgpack-ext", marks=using_msgpack),
    pytest.param("yaml", marks=using_pyyaml),
]


@pytest.mark.parametrize(
    "obj",
    [
        5,
        1.11111,
        "hello",
        "\u0394",
        numpy.random.rand(4),
        {"a": 5},
        {"a": 1.111111111111},
        {"a": "hello"},
        {
            "a": numpy.random.rand(4),
            "b": numpy.array(5.1111111),
            "c": numpy.array("hello"),
        },
        ["12345"],
        ["hello", "world"],
        [5, 123.234, "abcdé", "\u0394", "\U00000394"],
        [5, "B63", numpy.random.rand(4)],
        ["abcdé", {"a": numpy.random.rand(2), "b": numpy.random.rand(5)}],
        [
            numpy.array(3),
            numpy.arange(3, dtype=numpy.uint16),
            {"b": numpy.array(["a", "b"])},
        ],
    ],
)
@pytest.mark.parametrize("encoding", serialize_extensions)
def test_serialization(obj, encoding):
    new_obj = cmselemental.util.deserialize(
        cmselemental.util.serialize(obj, encoding=encoding), encoding=encoding
    )
    assert cmselemental.testing.compare_recursive(obj, new_obj)



import json
from pathlib import Path
from typing import Any, Dict, Optional, Set, Union

import numpy
from pydantic import BaseModel, BaseSettings

from ..testing import compare_recursive
from ..util import deserialize, serialize, yaml_import
from ..util.autodocs import AutoPydanticDocGenerator
from ..util.decorators import classproperty

cmsschema_draft = "http://json-schema.org/draft-07/schema#"

__all__ = ["ProtoModel", "AutodocBaseSettings"]


class ProtoModel(BaseModel):
    class Config:
        allow_mutation: bool = False
        extra: str = "forbid"
        json_encoders: Dict[str, Any] = {numpy.ndarray: lambda v: v.flatten().tolist()}
        serialize_default_excludes: Set = set()
        serialize_skip_defaults: bool = False
        force_skip_defaults: bool = False

        def schema_extra(schema, model):
            # below addresses the draft issue until https://github.com/samuelcolvin/pydantic/issues/1478 .
            schema["$schema"] = cmsschema_draft

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls.__doc__ = AutoPydanticDocGenerator(cls, always_apply=True)

    def __repr__(self):
        return f'{self.__repr_name__()}({self.__repr_str__(", ")})'

    def __str__(self):
        return f'{self.__repr_name__()}({self.__repr_str__(", ")})'

    @classproperty
    def default_schema_name(cls) -> Union[str, None]:
        """Returns default schema name if found."""
        try:
            return cls.schema()["properties"]["schema_name"]["default"]
        except Exception:
            return None

    @classmethod
    def parse_raw(cls, data: Union[bytes, str], *, encoding: str = None) -> "ProtoModel":  # type: ignore
        """
        Parses raw string or bytes into a Model object.
        Parameters
        ----------
        data : Union[bytes, str]
            A serialized data blob to be deserialized into a Model.
        encoding : str, optional
            The type of the serialized array, available types are: {'json', 'json-ext', 'msgpack-ext', 'pickle'}
        Returns
        -------
        Model
            The requested model from a serialized format.
        """

        if encoding is None:
            if isinstance(data, str):
                encoding = "json"  # Choose JSON over YAML by default
            elif isinstance(data, bytes):
                encoding = "msgpack-ext"
            else:
                raise TypeError(
                    "Input is neither str nor bytes, please specify an encoding."
                )

        if encoding.endswith(("json", "javascript", "pickle")):
            return super().parse_raw(data, content_type=encoding)
        elif encoding in ["msgpack-ext", "json-ext", "yaml"]:
            obj = deserialize(data, encoding)
        else:
            raise TypeError(f"Content type '{encoding}' not understood.")

        return cls.parse_obj(obj)

    @classmethod
    def parse_file(cls, path: Union[str, Path], *, encoding: str = None) -> "ProtoModel":  # type: ignore
        """Parses a file into a Model object.
        Parameters
        ----------
        path : Union[str, Path]
            The path to the file.
        encoding : str, optional
            The type of the files, available types are: {'json', 'msgpack', 'pickle', 'hdf5'}. Attempts to
            automatically infer the file type from the file extension if None.
        Returns
        -------
        Model
            The requested model from a file format.
        """
        path = Path(path)

        if encoding is None:
            if path.suffix in [".json", ".js"]:
                encoding = "json"
            elif path.suffix in [".yaml", ".yml"]:
                encoding = "yaml"
            elif path.suffix in [".msgpack"]:
                encoding = "msgpack-ext"
            elif path.suffix in [".pickle"]:
                encoding = "pickle"
            elif path.suffix in [".yaml", ".yml"]:
                encoding = "yaml"
            elif path.suffix in [".hdf5", ".h5"]:
                encoding = "hdf5"
            else:
                raise TypeError(
                    "Could not infer `encoding`, please provide a `encoding` for this file."
                )
        if encoding == "yaml":
            return cls.parse_raw(path.read_text(), encoding=encoding)
        elif encoding in ("hdf5", "h5"):
            from ..util import hdf

            return cls.parse_obj(hdf.read_file(path))
        return cls.parse_raw(path.read_bytes(), encoding=encoding)

    def write_file(
        self,
        path: Union[str, Path],
        *,
        encoding: str = None,
        mode: str = "w",
        **kwargs: Optional[Dict[str, Any]],
    ):
        """Write a Model to an output file.
        Parameters
        ----------
        path : Union[str, Path]
            The path to the file.
        encoding : str, optional
            The type of the files, available types are: {'json', 'msgpack', 'pickle', 'hdf5'}. Attempts to
            automatically infer the file type from the file extension if None.
        mode : str, optional
            An optional string that specifies the mode in which the file is written. Overwrites existing
            file by default (mode='w'). For appending to existing file, set mode='a'.
        **kwargs: Dict[str, Any], optional
            Additional keyword arguments passed to self.dict(), allows which fields to include, exclude, etc.
        """
        encoding = encoding or Path(path).suffix[1:]

        if encoding in ["json", "js", "yaml", "yml"]:
            stringified = self.serialize(encoding=encoding, **kwargs)
            with open(path, mode) as fp:
                fp.write(stringified)
        elif encoding in ["hdf5", "h5"]:
            from ..util import hdf

            hdf.write_file(path, data=self.dict(**kwargs), mode=mode)

    def dict(
        self, *, ser_kwargs: Dict[str, Any] = {}, **kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Returns object fields as a dictionary.
        Parameters
        ----------
        ser_kwargs: Optional[Dict[str, Any]]
            Additional keyword arguments to pass to serialize.
        **kwargs: Optional[Dict[str, Any]]
            Additional keyword arguments, allow which fields to include, exclude, etc.
        Returns
        -------
        Dict[str, Any]
            Fields as a dictionary.
        """
        encoding = kwargs.pop("encoding", None)

        kwargs["exclude"] = (
            kwargs.get("exclude", None) or set()
        ) | self.__config__.serialize_default_excludes  # type: ignore
        kwargs.setdefault("exclude_unset", self.__config__.serialize_skip_defaults)  # type: ignore
        if self.__config__.force_skip_defaults:  # type: ignore
            kwargs["exclude_unset"] = True

        data = super().dict(**kwargs)

        if encoding is None:
            return data
        elif encoding == "json":
            return json.loads(serialize(data, encoding=encoding, **ser_kwargs))
        elif encoding == "yaml":
            yaml = yaml_import(raise_error=True)
            return yaml.safe_load(serialize(data, encoding=encoding, **ser_kwargs))
        else:
            raise KeyError(
                f"Unknown encoding type '{encoding}', valid encoding types: 'json', 'yaml'."
            )

    def serialize(
        self,
        encoding: str,
        *,
        include: Optional[Set[str]] = None,
        exclude: Optional[Set[str]] = None,
        exclude_unset: Optional[bool] = None,
        exclude_defaults: Optional[bool] = None,
        exclude_none: Optional[bool] = None,
        **kwargs: Optional[Dict[str, Any]],
    ) -> Union[bytes, str]:
        """Generates a serialized representation of the model
        Parameters
        ----------
        encoding : str
            The serialization type, available types are: {'json', 'json-ext', 'msgpack-ext'}
        include : Optional[Set[str]], optional
            Fields to be included in the serialization.
        exclude : Optional[Set[str]], optional
            Fields to be excluded in the serialization.
        exclude_unset : Optional[bool], optional
            If True, skips fields that have default values provided.
        exclude_defaults: Optional[bool], optional
            If True, skips fields that have set or defaulted values equal to the default.
        exclude_none: Optional[bool], optional
            If True, skips fields that have value ``None``.
         **kwargs: Optional[Dict[str, Any]]
            Additional keyword arguments to pass to serialize.
        Returns
        -------
        Union[bytes, str]
            The serialized model.
        """

        fdargs = {}
        if include:
            fdargs["include"] = include
        if exclude:
            fdargs["exclude"] = exclude
        if exclude_unset:
            fdargs["exclude_unset"] = exclude_unset
        if exclude_defaults:
            fdargs["exclude_defaults"] = exclude_defaults
        if exclude_none:
            fdargs["exclude_none"] = exclude_none

        data = self.dict(**fdargs)

        if encoding == "js":
            encoding = "json"
        elif encoding == "yml":
            encoding = "yaml"

        return serialize(data, encoding=encoding, **kwargs)

    def json(self, **kwargs):
        # Alias JSON here from BaseModel to reflect dict changes
        return self.serialize("json", **kwargs)

    def yaml(self, **kwargs):
        return self.serialize("yaml", **kwargs)

    def compare(self, other: Union["ProtoModel", BaseModel], **kwargs) -> bool:
        """Compares the current object to the provided object recursively.
        Parameters
        ----------
        other : Model
            The model to compare to.
        **kwargs
            Additional kwargs to pass.
        Returns
        -------
        bool
            True if the objects match.
        """
        return compare_recursive(self, other, **kwargs)


class AutodocBaseSettings(BaseSettings):
    def __init_subclass__(cls) -> None:
        cls.__doc__ = AutoPydanticDocGenerator(cls, always_apply=True)



"""
Unit and regression test for the cmselemental package.
"""

# Import package, test suite, and other packages as needed
import cmselemental
import sys


def test_cmselemental_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "cmselemental" in sys.modules



import pytest

from cmselemental.models import (
    ComputeError,
    FailedOperation,
    InputProc,
    OutputProc,
    ProtoModel,
    Provenance,
)


@pytest.mark.skip(reason="no way of currently testing this")
def test_repr_provenance(request):

    prov = Provenance(creator="cmsel", version="v0.3.2")
    drop_qcsk(prov, request.node.name)

    assert "cmsel" in str(prov)
    assert "cmsel" in repr(prov)


def test_repr_compute_error():
    ce = ComputeError(error_type="random_error", error_message="this is bad")

    assert "random_error" in str(ce)
    assert "random_error" in repr(ce)


def test_repr_failed_op():
    fail_op = FailedOperation(
        error=ComputeError(error_type="random_error", error_message="this is bad")
    )
    assert (
        str(fail_op)
        == """FailedOperation(error=ComputeError(error_type='random_error', error_message='this is bad'))"""
    )


def test_repr_proc_input():

    opt = InputProc(
        **{
            "schema_name": "some_schema",
            "schema_version": 2,
            "id": "some_id",
            "engine": "some_engine",
            "engine_version": "1.0.0",
        }
    )

    assert "provenance" in str(opt)
    assert "schema_name" in str(opt)
    assert "schema_version" in str(opt)


def test_repr_proc_output():

    opt = OutputProc(
        **{
            "schema_name": "my_schema",
            "schema_version": 1,
            "stdout": "stdout",
            "stderr": "stderr",
            "success": False,
        }
    )

    assert "provenance" in str(opt)
    assert "schema_name" in str(opt)
    assert "schema_version" in str(opt)


def test_model_custom_repr():
    class Model(ProtoModel):
        a: int

        def __repr__(self) -> str:
            return "Hello world!"

    m = Model(a=5)
    assert repr(m) == "Hello world!"
    assert "Model(" in str(m)

    class Model2(ProtoModel):
        a: int

        def __str__(self) -> str:
            return "Hello world!"

    m = Model2(a=5)
    assert "Model2(" in repr(m)
    assert str(m) == "Hello world!"



from typing import Any, Dict, Optional
from pydantic import Field

from ..extras import provenance_stamp
from .base import ProtoModel
from .common import (
    ComputeError,
    Provenance,
)

__all__ = ["InputProc", "OutputProc"]


class InputProc(ProtoModel):
    id: Optional[str] = None
    hash_index: Optional[str] = None
    schema_name: str = Field(  # type: ignore
        ...,
        description=("The schema specification to which this model conforms."),
    )
    schema_version: int = Field(  # type: ignore
        ...,
        description="The version number of ``schema_name`` to which this model conforms.",
    )
    keywords: Optional[Dict[str, Any]] = Field(
        {}, description="Procedure specific keywords to be used."
    )
    provenance: Optional[Provenance] = Field(
        Provenance(**provenance_stamp(__name__)), description=str(Provenance.__doc__)
    )
    engine: Optional[str] = Field(
        None,
        description="Engine name to use in the procedure e.g. OpenMM.",
    )
    engine_version: Optional[str] = Field(
        None, description="Supported engine version. e.g. '>=3.4.0'."
    )
    extras: Optional[Dict[str, Any]] = Field(
        {}, description="Extra fields that are not part of the schema."
    )


class OutputProc(ProtoModel):
    proc_input: Optional[InputProc] = None
    schema_name: str = Field(  # type: ignore
        ...,
        description=("The schema specification to which this model conforms."),
    )
    schema_version: int = Field(  # type: ignore
        ...,
        description="The version number of ``schema_name`` to which this model conforms.",
    )
    stdout: Optional[str] = Field(
        None, description="The standard output of the program."
    )
    stderr: Optional[str] = Field(
        None, description="The standard error of the program."
    )
    warnings: Optional[str] = Field(None, description="Warning messages.")
    log: Optional[str] = Field(None, description="Logging info.")
    success: bool = Field(
        ...,
        description="The success of a given programs execution. If False, other fields may be blank.",
    )
    error: Optional[ComputeError] = Field(None, description=str(ComputeError.__doc__))
    provenance: Optional[Provenance] = Field(
        Provenance(**provenance_stamp(__name__)), description=str(Provenance.__doc__)
    )
    extras: Optional[Dict[str, Any]] = Field(
        {}, description="Extra fields that are not part of the schema."
    )



import pytest
import numpy
import cmselemental
from pathlib import Path


using_h5py = pytest.mark.skipif(
    cmselemental.util.which_import("h5py", return_bool=True) is False,
    reason="Not detecting module h5py. Install package if necessary and add to envvar PYTHONPATH",
)


@using_h5py
@pytest.mark.parametrize(
    "obj",
    [
        {
            "a": numpy.random.rand(4),
            "b": numpy.array(5.1111111),
            "c": numpy.array("hello"),
            "abcdé": {"a": numpy.random.rand(2), "b": numpy.random.rand(5)},
            "nested": [(1, 0), (0, 1), (2, 2)],
        },
        {
            "a": numpy.array([(0, 1, 2), (2, 3, 4)], dtype="i8,i8,i8"),
            "b": numpy.array([("ALA", 1.0, 2.0), ("FF", 3.0, 4.0)], dtype="U3,f8,f8"),
        },
    ],
)
def test_write(obj):
    from cmselemental.util.hdf import write_file

    path_to_file = Path("filename.h5")
    write_file(path_to_file.name, data=obj)
    assert path_to_file.is_file()
    path_to_file.unlink()



import json
from typing import Any, Union, Dict, Optional

import numpy as np
from pydantic.json import pydantic_encoder

from .importing import which_import, yaml_import

try:
    import msgpack
except ModuleNotFoundError:
    pass

_msgpack_which_msg = "Please install via `conda install msgpack-python`."

## MSGPackExt


def msgpackext_encode(obj: Any) -> Any:
    """
    Encodes an object using pydantic and NumPy array serialization techniques suitable for msgpack.
    Parameters
    ----------
    obj : Any
        Any object that can be serialized with pydantic and NumPy encoding techniques.
    Returns
    -------
    Any
        A msgpack compatible form of the object.
    """

    # First try pydantic base objects
    try:
        return pydantic_encoder(obj)
    except TypeError:
        pass

    if isinstance(obj, np.ndarray):
        if obj.shape:
            data = {
                b"_nd_": True,
                b"dtype": obj.dtype.str,
                b"data": np.ascontiguousarray(obj).tobytes(),
            }
            if len(obj.shape) > 1:
                data[b"shape"] = obj.shape
            return data

        else:
            # Converts np.array(5) -> 5
            return obj.tolist()

    return obj


def msgpackext_decode(obj: Any) -> Any:
    """
    Decodes a msgpack objects from a dictionary representation.
    Parameters
    ----------
    obj : Any
        An encoded object, likely a dictionary.
    Returns
    -------
    Any
        The decoded form of the object.
    """

    if b"_nd_" in obj:
        arr = np.frombuffer(obj[b"data"], dtype=obj[b"dtype"])
        if b"shape" in obj:
            arr.shape = obj[b"shape"]

        return arr

    return obj


def msgpackext_dumps(data: Any, **kwargs: Optional[Dict[str, Any]]) -> bytes:
    """Safe serialization of a Python object to msgpack binary representation using all known encoders.
    For NumPy, encodes a specialized object format to encode all shape and type data.
    Parameters
    ----------
    data : Any
        A encodable python object.
    **kwargs : Optional[Dict[str, Any]], optional
        Additional keyword arguments to pass to the constructor.
    Returns
    -------
    bytes
        A msgpack representation of the data in bytes.
    """
    which_import("msgpack", raise_error=True, raise_msg=_msgpack_which_msg)
    use_bin_type = kwargs.pop("use_bin_type", True)

    return msgpack.dumps(
        data, default=msgpackext_encode, use_bin_type=use_bin_type, **kwargs
    )


def msgpackext_loads(data: bytes, **kwargs: Dict[str, Any]) -> Any:
    """Deserializes a msgpack byte representation of known objects into those objects.
    Parameters
    ----------
    data : bytes
        The serialized msgpack byte array.
    **kwargs : Optional[Dict[str, Any]], optional
        Additional keyword arguments to pass to the constructor.
    Returns
    -------
    Any
        The deserialized Python objects.
    """
    which_import("msgpack", raise_error=True, raise_msg=_msgpack_which_msg)
    raw = kwargs.pop("raw", False)
    return msgpack.loads(data, object_hook=msgpackext_decode, raw=raw, **kwargs)


## JSON Ext


class JSONExtArrayEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        try:
            return pydantic_encoder(obj)
        except TypeError:
            pass

        if isinstance(obj, np.ndarray):
            if obj.shape:
                data = {
                    "_nd_": True,
                    "dtype": obj.dtype.str,
                    "data": np.ascontiguousarray(obj).tobytes().hex(),
                }
                if len(obj.shape) > 1:
                    data["shape"] = obj.shape
                return data

            else:
                # Converts np.array(5) -> 5
                return obj.tolist()

        return json.JSONEncoder.default(self, obj)


def jsonext_decode(obj: Any) -> Any:

    if "_nd_" in obj:
        arr = np.frombuffer(bytes.fromhex(obj["data"]), dtype=obj["dtype"])
        if "shape" in obj:
            arr.shape = obj["shape"]

        return arr

    return obj


def jsonext_dumps(data: Any, **kwargs: Optional[Dict[str, Any]]) -> str:
    """Safe serialization of Python objects to JSON string representation using all known encoders.
    The JSON serializer uses a custom array syntax rather than flat JSON lists.
    Parameters
    ----------
    data : Any
        A encodable python object.
    **kwargs : Optional[Dict[str, Any]], optional
        Additional keyword arguments to pass to the constructor
    Returns
    -------
    str
        A JSON representation of the data.
    """

    return json.dumps(data, cls=JSONExtArrayEncoder, **kwargs)


def jsonext_loads(data: Union[str, bytes]) -> Any:
    """Deserializes a json representation of known objects into those objects.
    Parameters
    ----------
    data : str or bytes
        The byte-serialized JSON blob.
    Returns
    -------
    Any
        The deserialized Python objects.
    """

    return json.loads(data, object_hook=jsonext_decode)


## JSON


class JSONArrayEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        try:
            return pydantic_encoder(obj)
        except TypeError:
            pass

        if isinstance(obj, np.ndarray):
            if obj.shape:
                return obj.ravel().tolist()
            else:
                return obj.tolist()

        return json.JSONEncoder.default(self, obj)


def json_dumps(data: Any, **kwargs: Optional[Dict[str, Any]]) -> str:
    """Safe serialization of a Python dictionary to JSON string representation using all known encoders.
    Parameters
    ----------
    data : Any
        A encodable python object.
    **kwargs : Optional[Dict[str, Any]], optional
        Additional keyword arguments to pass to the constructor
    Returns
    -------
    str
        A JSON representation of the data.
    """

    return json.dumps(data, cls=JSONArrayEncoder, **kwargs)


def json_loads(data: str) -> Any:
    """Deserializes a json representation of known objects into those objects.
    Parameters
    ----------
    data : str
        The serialized JSON blob.
    Returns
    -------
    Any
        The deserialized Python objects.
    """

    # Doesn't hurt anything to try to load JSONext as well
    return json.loads(data, object_hook=jsonext_decode)


## YAML


def yaml_encode(
    dumper: Union["yaml.dumper.SafeDumper", "yaml.dumper.RoundTripDumper"],
    obj: np.ndarray,
) -> "yaml.nodes.Node":
    """
    Encodes a NumPy ndarray for YAML.
    Parameters
    ----------
    dumper: yaml.dumper.SafeDumper or ruamel.yaml.dumper.RoundTripDumper
        A PyYAML SafeDumper or ruamel RoundTripDumper object
    obj : np.ndarray
        A NumPy ndarray object
    Returns
    -------
    yaml.nodes.Node
        A YAML node representing the object.
    """

    if obj.shape:
        return dumper.represent_list(obj.ravel().tolist())
    else:
        # Converts np.array(5) -> 5
        return dumper.represent_data(obj.tolist())


def yaml_safe_dump(data, stream=None, sort_keys=False, **kwargs):
    """Mimics yaml.safe_dump with support for numpy.ndarray encoding. If stream is None, return
    the produced string instead. Order is preserved by default."""

    yaml = yaml_import(raise_error=True)

    if yaml.__name__ == "ruamel.yaml":

        class SafeDumper(yaml.RoundTripDumper):
            ...

        SafeDumper.add_representer(np.ndarray, yaml_encode)
        return yaml.dump(data, stream=stream, Dumper=SafeDumper, **kwargs)
    else:

        class SafeDumper(yaml.SafeDumper):
            ...

        SafeDumper.add_representer(np.ndarray, yaml_encode)
        return yaml.dump(
            data, stream=stream, Dumper=SafeDumper, sort_keys=sort_keys, **kwargs
        )


def yaml_dump(data: Any, **kwargs: Optional[Dict[str, Any]]) -> str:
    """Safe serialization of a Python dictionary to YAML string representation.
    Parameters
    ----------
    data : Any
        An encodable python object.
    **kwargs : Optional[Dict[str, Any]], optional
        Additional keyword arguments to pass to the constructor
    Returns
    -------
    str
        A YAML representation of the data.
    """
    return yaml_safe_dump(data, **kwargs)


def yaml_load(data: str) -> Any:
    """Deserializes a yaml representation of known objects into those objects.
    Parameters
    ----------
    data : str
        The serialized YAML blob.
    Returns
    -------
    Any
        The deserialized Python objects.
    """

    yaml = yaml_import(raise_error=True)
    return yaml.safe_load(data)


## Helper functions


def serialize(
    data: Any, encoding: str, **kwargs: Optional[Dict[str, Any]]
) -> Union[str, bytes]:
    """Encoding Python objects using the provided encoder.
    Parameters
    ----------
    data : Any
        A encodable python object.
    encoding : str
        The type of encoding to perform: {'json', 'json-ext', 'yaml', 'msgpack-ext'}
    **kwargs : Optional[Dict[str, Any]], optional
        Additional keyword arguments to pass to the constructors.
    Returns
    -------
    Union[str, bytes]
        A serialized representation of the data.
    """
    if encoding.lower() == "json":
        return json_dumps(data, **kwargs)
    elif encoding.lower() == "json-ext":
        return jsonext_dumps(data, **kwargs)
    elif encoding.lower() == "yaml":
        return yaml_dump(data, **kwargs)
    elif encoding.lower() == "msgpack-ext":
        return msgpackext_dumps(data, **kwargs)
    else:
        raise KeyError(
            f"Encoding '{encoding}' not understood, valid options: 'json', 'json-ext', 'yaml', 'msgpack-ext'"
        )


def deserialize(blob: Union[str, bytes], encoding: str) -> Any:
    """Encoding Python objects using .
    Parameters
    ----------
    blob : Union[str, bytes]
        The serialized data.
    encoding : str
        The type of encoding of the blob: {'json', 'json-ext', 'msgpack'}
    Returns
    -------
    Any
        The deserialized Python objects.
    """
    if encoding.lower() == "json":
        assert isinstance(blob, str)
        return json_loads(blob)
    elif encoding.lower() == "json-ext":
        assert isinstance(blob, (str, bytes))
        return jsonext_loads(blob)
    elif encoding.lower() == "yaml":
        assert isinstance(blob, str)
        return yaml_load(blob)
    elif encoding.lower() in ["msgpack", "msgpack-ext"]:
        assert isinstance(blob, bytes)
        return msgpackext_loads(blob)
    else:
        raise KeyError(
            f"Encoding '{encoding}' not understood, valid options: 'json', 'json-ext', 'msgpack-ext'"
        )



from typing import Callable
import functools


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


def require(pkg_name: str) -> Callable:
    """Returns a decorator function, ensures pkg_name is available and can be imported.
    Parameters
    ----------
    pkg_name: str
        Name of the package required.
    Returns
    -------
    deco_require: Callable
        Decorator function
    Raises
    ------
    ModuleNotFoundError
        When pkg_name is not found.
    Example:
    --------
    @require("some_pkg")
    def foo(...):
        ...
    """

    def deco_require(func):
        @functools.wraps(func)
        def inner_func(*args, **kwargs):
            if not which_import(pkg_name, return_bool=True):
                raise ModuleNotFoundError(f"Could not find or import {pkg_name}.")
            return func(*args, **kwargs)

        return inner_func

    return deco_require



from . import exceptions
from . import importing
from .importing import yaml_import, which_import, which
from . import serialization
from .serialization import serialize, deserialize
from . import autodocs
from . import decorators



import pytest
import cmselemental

qcel_found = cmselemental.util.importing.which_import("qcelemental")


@pytest.mark.skipif(not qcel_found, reason="qcelemental not installed")
def test_optim_qcel():
    from qcelemental.models.procedures import (
        OptimizationProtocols,
        QCInputSpecification,
    )
    from qcelemental.models import Molecule

    class OptimizationInput(cmselemental.models.InputProc):

        schema_name: str = "qcschema"
        schema_version: int = 2
        protocols: OptimizationProtocols = OptimizationProtocols()
        input_specification: QCInputSpecification = ...
        initial_molecule: Molecule = ...

        def __repr_args__(self) -> "ReprArgs":
            return [
                ("model", self.input_specification.model.dict()),
                ("molecule_hash", self.initial_molecule.get_hash()[:7]),
            ]

    opt = OptimizationInput(
        **{
            "input_specification": {"driver": "gradient", "model": {"method": "UFF"}},
            "initial_molecule": {"symbols": ["He"], "geometry": [0, 0, 0]},
        }
    )

    assert "molecule_hash" in str(opt)
    assert "molecule_hash" in repr(opt)



try:
    import pydantic
except ImportError:  # pragma: no cover
    raise ImportError(
        "Python module pydantic not found. Solve by installing it: "
        "`conda install pydantic -c conda-forge` or `pip install pydantic`"
    )

from .base import *
from .common import *
from .procedures import *



from typing import Dict, Any
import numpy
import json

try:
    import h5py
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Hdf5 functionality requires h5py. Solve by: pip install h5py."
    )

import collections
from contextlib import redirect_stdout

encoding = "utf-8"
str_encode = h5py.string_dtype(encoding=encoding)

numpy_h5py_dtypes = {
    "U": str_encode,
    "i": "int64",
    "f": "float64",
}


def h5tojson(h5filename: str, jsonfilename: str, mode: str = "w"):
    from h5json import Hdf5db
    from h5json import hdf5dtype
    from h5json.h5tojson.h5tojson import getTempFileName, DumpJson

    options = collections.namedtuple("options", ["D", "d", "filename"])(
        D=None, d=None, filename=h5filename
    )

    with Hdf5db(
        options.filename, dbFilePath=getTempFileName(), readonly=True, app_logger=None
    ) as db:
        dumper = DumpJson(db, app_logger=None, options=options)
        with open(jsonfilename, mode) as fileobj:
            with redirect_stdout(fileobj):
                dumper.dumpFile()


def jsontoh5(jsonfilename: str, h5filename: str):
    from h5json import Hdf5db
    from h5json.jsontoh5.jsontoh5 import Writeh5

    with open(jsonfilename, "r") as fileobj:
        stringified = json.load(fileobj)

    if "root" not in stringified:
        raise Exception("no root key in input file")
    root_uuid = stringified["root"]

    # create the file, will raise IOError if there's a problem
    Hdf5db.createHDF5File(h5filename)

    with Hdf5db(
        h5filename, root_uuid=root_uuid, update_timestamps=False, app_logger=None
    ) as db:
        h5writer = Writeh5(db, stringified)
        h5writer.writeFile()

    # open with h5py and remove the _db_ group
    # Note: this will delete any anonymous (un-linked) objects
    with h5py.File(h5filename, "a") as fileobj:
        if "__db__" in fileobj:
            del fileobj["__db__"]


def write_file(filename: str, data: Dict[str, Any], mode: str = "w", **kwargs):
    with h5py.File(filename, mode) as hdfobj:
        write_dict(hdfobj, data)


def read_file(filename: str, **kwargs) -> Dict[str, Any]:
    with h5py.File(filename, "r") as hdfobj:
        return read_dict(hdfobj, **kwargs)


def _get_dtype(data):
    """Returns data type for hdf5 datasets."""
    if isinstance(data, str):
        return str_encode
    elif isinstance(data, float):
        return "float64"
    elif isinstance(data, int):
        return "int64"
    elif isinstance(data, numpy.ndarray):
        dtype = numpy_h5py_dtypes.get(data.dtype.kind, None)
        if dtype is None and data.dtype.kind == "V":
            dtype = [
                (f"f{i}", numpy_h5py_dtypes.get(data.dtype[i].kind))
                for i in range(len(data.dtype))
            ]
            if None in dtype:
                raise NotImplementedError(f"Data type {data.dtype} not supported.")
        return dtype
    elif isinstance(data, (list, tuple)):
        if isinstance(data[0], tuple):
            return ",".join([_get_dtype(item) for item in data[0]])
        elif isinstance(data[0], list):
            raise NotImplementedError("List of lists not supported.")
        else:
            assert all(
                isinstance(item, int) for item in data
            ), "Only homogenous arrays supported."  # keep this? Can degrade performance
            return _get_dtype(data[0])  # assume homogenous list/tuple
    else:
        raise NotImplementedError(f"Data type {type(data)} not supported.")


def _wrap_homogenous_array(data):
    dtype = _get_dtype(data)
    return numpy.array(data, dtype=dtype)


def write_dict(
    hdfobj: "h5py._hl.files.File",
    data: Dict[str, Any],
    units_metadata: bool = True,
    **kwargs,
) -> None:
    """
    Writes a python dictionary to an HDF5 file. By default, any attribute that ends in '_units' is stored
    as metadata in the hdf5 file. Can be turned off by setting units_metadata=False.

    Parameters
    ----------
    hdfobj: h5py._hl.files.File
        The hdf file object to write data to.
    data: Dict[str, Any]
        The dictionary of data to write.
    units_metadata: bool
        Treat any key ending in '_units' as metadata.
    **kwargs: Optional[Dict[str, Any]], optional
        Any additional keywords to pass to the constructor.

    """
    for key, val in data.items():
        if isinstance(val, dict):
            grp = hdfobj.create_group(key)
            write_dict(grp, val)
        elif units_metadata:
            if not key.endswith("_units"):  # Deal with units later
                val = _wrap_homogenous_array(val)
                hdfobj.create_dataset(name=key, data=val)

    # Save units as metadata
    if units_metadata:
        for key, val in data.items():
            if key.endswith("_units"):
                array_name, _ = key.split("_units")
                hdfobj[array_name].attrs[key] = val


def read_dict(hdfobj: "h5py._hl.files.File", **kwargs) -> Dict[str, Any]:
    """
    Converts an hdf file object to a python dictionary.

    Parameters
    ----------
    hdfobj: h5py._hl.files.File
        The hdf file object to read data from.
    **kwargs: Optional[Dict[str, Any]], optional
        Any additional keywords to pass to the constructor.
    Returns
    -------
    Dict[str, Any]
        A python dictionary that stores HDF5 data.

    """

    data = {}
    for key in hdfobj.keys():
        if isinstance(hdfobj[key], h5py.Group):
            data[key] = read_dict(hdfobj[key])
        elif isinstance(hdfobj[key], h5py.Dataset):
            data[key] = hdfobj[key][()]

            # For MMEl, store key_units as metadata
            if key + "_units" in hdfobj[key].attrs.keys():
                data[key + "_units"] = hdfobj[key].attrs[key + "_units"]

            if isinstance(data[key], bytes):
                data[key] = data[key].decode()
            elif isinstance(data[key], numpy.ndarray):
                if data[key].dtype.char == "O":
                    if len(data[key].dtype) < 2:  # homogenous array
                        data[key] = data[key].astype("U")  # unicode default in py3
                    else:
                        pass  # do nothing with homogenous array
        else:
            raise ValueError(f"Data type not understood: {hdfobj[key]}")
    return data
